/// @file ActionSet.hpp
/// @author Franck Dary
/// @version 1.0
/// @date 2018-08-06

#ifndef ACTIONSET__H
#define ACTIONSET__H

#include <vector>
#include <map>
#include "Action.hpp"

/// @brief A set of Action.
///
/// Each Classifier has an ActionSet, it represents the Action that the Classifier is able to predict.
class ActionSet
{
  public :

  /// @brief The name of this ActionSet.
  std::string name;
  /// @brief The Action contaiend by this ActionSet.
  std::vector<Action> actions;
  /// @brief Map action's name to their index inside the vector actions.
  std::map<std::string, int> str2index;
  /// @brief Whether or not this ActionSet is dynamic.
  ///
  /// When an ActionSet is dynamic, new Action can be added to it at any time.\n
  /// A Classifier of type Prediction cannot have a dynamic ActionSet.
  bool isDynamic;
  /// @brief Whether or not this ActionSet have a default action.
  ///
  /// If it is the case, the default action is the first element of the vector actions.\n When getAction is called with an unknown name, the default action is returned.\n If an ActionSet is dynamic, it cannot have a default action.
  bool hasDefaultAction;
  /// @brief Index of the default action, if it has one.
  int defaultActionIndex;

  public :

  /// @brief Construct a new ActionSet from a file.
  ///
  /// @param filename The name of the file containing the ActionSet.
  /// @param isDynamic Whether or not this ActionSet is dynamic.
  ActionSet(const std::string & filename, bool isDynamic);
  /// @brief Print this ActionSet for debug purposes.
  ///
  /// @param output Where to print.
  void printForDebug(FILE * output);
  /// @brief Get the index of an action given its name.
  ///
  /// This is useful to map each action to an integer, this integer will be the index of the corresponding output neuron of the neural network of the Classifier.
  /// @param name The name of the Action.
  ///
  /// @return The index of the Action in the vector actions.
  int getActionIndex(const std::string & name);
  /// @brief Get the name of the Action given its index.
  ///
  /// This is useful to map each action to an integer, this integer will be the index of the corresponding output neuron of the neural network of the Classifier.
  /// @param actionIndex The index of the Action in the vector actions.
  ///
  /// @return The name of the Action.
  std::string getActionName(int actionIndex);
  /// @brief Get a pointer to the Action object given its name.
  ///
  /// @param name The name of the Action.
  ///
  /// @return A pointer to the Action.
  Action * getAction(const std::string & name);
  /// @brief Get a pointer to the default Action object.
  ///
  /// @return A pointer to the default Action.
  Action * getDefaultAction();
  /// @brief Get the number of actions contained in this set.
  ///
  /// @return The number of actions in this set.
  unsigned int size();
};

#endif
